/*****************************************************************************
 * Copyright (c) 2020, rrcc FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#pragma once

#include "Instruction.h"

namespace Instruction {
    class Calc : public Instruction {
    public:

        enum class Operation : uint8_t {
            AND = 0,
            OR = 1,
            ADD = 2,
            SUB = 3,
            LSHIFT = 4,
            RSHIFT = 5,
            ARSHIFT = 6,
            UNDEFINED = 7
        };

        Calc() {

        }

        Calc(Operation operation, uint8_t addressOffset, uint32_t addressDestination, uint32_t addressSource,
             uint32_t addressTarget) : operation(operation), addressOffset(addressOffset),
                                       addressDestination(addressDestination), addressSource(addressSource),
                                       addressTarget(addressTarget) {}

        CommandOp getCommand() const override {
            return CommandOp::CALC;
        };


        std::vector<uint8_t> toBytes() const override;

        size_t getByteSize() const override{
            return 4 * 3;
        }


        Operation operation;
        uint8_t addressOffset;
        uint32_t addressDestination;
        uint32_t addressSource;
        uint32_t addressTarget;
    };
}